/* knnfp.c: k-Nearest Neighbor on Feature Projections (KNNFP)
   that considers each feature separately.
   Each instance is stored by its projections on features.
   Prediction of a feature is a vector of class counts 
   Weights are used if provided in the domain_name.info file 
   ******************************************************************/

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>
#include "knnfp.h"

int k=1;
int No_Features, No_Classes;
int No_Trainers, No_Testers;
int linear[MAX_FEATURES];
int verbosity=0;
int className[MAX_CLASSES], classIndex[MAX_CLASSES];
char domainName[MAX_NAME], logFileName[MAX_NAME], resFileName[MAX_NAME]; 

INSTANCE trainData[MAX_TRAIN], testData[MAX_TEST];
int classCount[MAX_CLASSES]; 
float vote[MAX_CLASSES]; /* votes are weighted by feature weights */
POINT sortedTrain[MAX_FEATURES][MAX_TRAIN]; /* sorted Training Data */
int kBag[MAX_CLASSES];  /* to keep the k nearest neighbor for a feature */
float weight[MAX_FEATURES]; /* feature weights */

FILE *resultf, *logf;

struct timeb start_train, end_train, end_test;
int train_time, test_time;

/*****************************************************************************/
int
time_diff(struct timeb finish, struct timeb start)
/* time difference in miliseconds */
{ int time_in_sec, time_in_msec;

  time_in_sec = finish.time - start.time;
  time_in_msec = finish.millitm - start.millitm;
  return(time_in_sec * 1000 + time_in_msec);
} /* time_diff */

/* ------------------------------------------------------------------------ */
void 
parse_options(int argc, char *argv[])
{ int i=2;

  strcpy(domainName, argv[1]);
  while (i<argc) {
    if (strcmp(argv[i],"-k")==0)
      k = atoi(argv[++i]);
    else if(strcmp(argv[i], "-v")==0 ) {
      sscanf(argv[++i], "%d", &verbosity);
      sprintf(logFileName,"%s.log.knnfp", domainName);
      logf = fopen(logFileName,"w");
      fprintf(logf,"Classifier: knnfp\n");
      fflush(logf);
    }
    else printf ("Unrecognized command line argument: %s\n", argv[i]);
    i++;
  }
} /* parse_options */

/*****************************************************************************/
static int 
VALUEcompare(i,j)
POINT *i, *j;
{
  /* UNKNOWN values are treated as INFINITE */
  if ((*i).status == UNKNOWN)  return (1);
  if ((*j).status == UNKNOWN)  return (-1);
  if ((*i).value > (*j).value) return (1);
  if ((*i).value < (*j).value) return (-1);
  return(0);
} /* VALUEcompare */

/*****************************************************************************/
/* Training involves only sorting the training data.                         */
void 
train()
{
  FEATURE f;
  int i;

  for (i=0; i<No_Trainers; i++)
    for (f=0; f<No_Features; f++) {
      sortedTrain[f][i].status = trainData[i].status[f];
      sortedTrain[f][i].value  = trainData[i].value[f];
      sortedTrain[f][i].class  = trainData[i].class;
    }
  for(f=0; f<No_Features; f++) 
    qsort((char *)sortedTrain[f], No_Trainers, sizeof(POINT), VALUEcompare);
} /* train() */
      
/*****************************************************************************/
/* returns distance between training data point and test value on f          */
float 
distance (i, v, f)
int i;    /* i is index of the point in sortedTrain */
VALUE v;  /* value of the test instance */
FEATURE f; 
{
  if (sortedTrain[f][i].status == UNKNOWN) return(INFINITE);
  if (!linear[f]) return((float)(sortedTrain[f][i].value != v));
  if (sortedTrain[f][i].value > v)
    return(sortedTrain[f][i].value - v);
  return(v - sortedTrain[f][i].value) ;
} /* distance */

/*****************************************************************************/
int 
nearest(FEATURE f, VALUE v, int min, int max)
{
  int mid;
   
  if (sortedTrain[f][min].status == UNKNOWN ){
    fprintf(stderr,"feature[%d] seems to be all UNKNOWN", f);
    exit(0);
  }
  if ((max-min)==1)
    if (sortedTrain[f][max].status == UNKNOWN)
      return(min);
    else if ((sortedTrain[f][min].value - v) < (sortedTrain[f][max].value - v))
      return(min);
    else return(max);
  
  mid = (min+max)/2;
  if(sortedTrain[f][mid].status == UNKNOWN)
    return(nearest(f, v, min, mid));
  if(sortedTrain[f][mid].value == v)
    return(mid);
  if(sortedTrain[f][mid].value < v)
    return(nearest(f, v, mid, max));
  return(nearest(f, v, min, mid));
} /* nearest(f, v, min, max) */

/*****************************************************************************/
void 
fill_kBag(f, v)
FEATURE f;
VALUE v;
{
  int up, down;
  int n;			/* n: neighbor */

  /*  Assumes: k < No of KNOWN Trainers */
  up = down = nearest(f, v, 0, No_Trainers-1);
  kBag[sortedTrain[f][up++].class]++; down--;

  for (n=1; n<k; n++) {	   /* find k nearest neighbors in feature f */
    if (up < No_Trainers) {
      if (down > -1)		/* can move both directions */
	if (distance (up, v, f) < distance (down, v, f)) 
	  kBag[sortedTrain[f][up++].class]++;
	else kBag[sortedTrain[f][down--].class]++;
      else			/* can't move down */
	kBag[sortedTrain[f][up++].class]++;
    }
    else			/* can't move up */
      kBag[sortedTrain[f][down--].class]++;
  }
} /* fill_kBag */

/*****************************************************************************/
float 
knnfp()
{
  int test;
  FEATURE f;
  CLASS c, prediction;
  int correct_count=0;
  
  for (test=0; test < No_Testers; test++) { 
    for (c=0; c <= No_Classes; vote[c++]=0.0); /* initalize for voting */
    for (f=0; f < No_Features; f++) { 
      /* each feature makes its individual contribution to voting */
      if (testData[test].status[f] == KNOWN) {
	for (c=0; c <= No_Classes; kBag[c++]=0); /* initialize kBag */
	/* find k nearest neighbor to the test value on feature f */
	fill_kBag(f, testData[test].value[f]);

	if(verbosity > 3) {
	  fprintf(logf,"bag from feature %d: [ ", f);
	  for (c=0; c<=No_Classes; c++)
	    fprintf(logf, "%d/%d ", className[c], kBag[c]);
	  fprintf(logf,"]\n");
	}

	for (c=1; c <= No_Classes; c++) /* add weighted kBag to vote vactor */
	  vote[c] += weight[f] * kBag[c];
      } /* for known values */
    } /* for each feature */

    if (verbosity > 2) {
      fprintf(logf, "\nVotes: ");
      for (c=1; c<=No_Classes; c++)
	fprintf(logf, "Class %d: %5.3f ", className[c], vote[c]);
      fprintf(logf, "\n");
    }

    /* Voting */  
    prediction = UNDETERMINED;
    for(c=1; c <= No_Classes; c++) 
      if (vote[c] > vote[prediction]) prediction = c;

    if (verbosity>1) 
      fprintf(logf, "Test instance: %d, Prediction: %d, Actual class: %d\n", 
	      test, className[prediction], className[testData[test].class]);
    if( prediction == testData[test].class )
      correct_count++;
  } /* for all test instances */

  return((float)correct_count/No_Testers);
} /* knnfp */

/*****************************************************************************/
void 
initialize ()
{ 
  int i;
  CLASS c;
  FEATURE f;

  for (f=0; f<MAX_FEATURES; f++) weight[f] = 1.0;          /* default weight */
  read_info();
  if (verbosity > 0) {
    fprintf(logf,"Domanin: %s, k: %d\n", domainName, k);
    fprintf(logf,"No_Features: %d\n", No_Features);
    fprintf(logf,"No_Classes: %d\n", No_Classes);
  }

  No_Trainers = read_train_data(); /* initialize trainData from domain.train */
  for (i=0; i<No_Trainers; i++)
    ++classCount[trainData[i].class];

  No_Testers = read_test_data();   /* initialize testData  from domain.test  */

  if (verbosity > 3) {
    fprintf(logf, "train data:\n");
    log_data(trainData, No_Trainers); 
    fprintf(logf, "test data:\n");
    log_data(testData, No_Testers); 
  }
  if (verbosity > 1) {
    fprintf(logf,"No_Trainers: %d\n", No_Trainers);
    fprintf(logf,"No_Testers: %d\n", No_Testers);
    for (c=1; c <= No_Classes; c++) 
      fprintf(logf, "classCount[%d]:%d ", className[c], classCount[c]);
    fprintf(logf, "\n");
    for (f=0; f<No_Features; f++)
      fprintf(logf, "Weight[%d]=%5.3f ", f, weight[f]);
    fprintf(logf, "\n");
  }

  if (k > No_Trainers) {
    fprintf(logf,"k (%d) > Number of trainers: (%d) !!! EXIT !!!\n", 
	    k, No_Trainers);
    exit(1);
  }
} /* initialize () */

/*****************************************************************************/
main(int argc, char *argv[])
{
  float accuracy;
 
  if (argc == 1) {
    printf("k-Nearest Neighbor Classifes on Feature Projections\n");
    printf("knnfp domain_name [-k k] [-v v]\n");
    printf("k : number of neighbors to be taken (default is 1),\n"); 
    printf("v : degree of the verbosity (default is 0).\n");
    printf("Reads training data from domain_name.train,\n");
    printf("and test data from domain_name.test.\n");
    printf("Returns accuracy and run time in the file domain_name.res.\n");
    exit(1);
  }
  else
    parse_options(argc,argv);

  tzset(); /* set time zone */

  initialize();			/* read train and test data */

  ftime(&start_train);
  train (); 
  ftime(&end_train);

  accuracy = knnfp(); 
  ftime(&end_test);

  train_time = time_diff (end_train, start_train);
  test_time  = time_diff (end_test, end_train);

  sprintf(resFileName,"%s.res.knnfp", domainName);
  resultf = fopen(resFileName, "w");
  fprintf (resultf, "Accuracy: %5.3f Train: %d Test: %d msec.\n", 
	   accuracy, train_time, test_time);

  if (verbosity>0)
    fprintf (logf,"Accuracy: %5.3f Train: %d Test: %d msec.\n", 
	   accuracy, train_time, test_time); 
  fclose(resultf);
  if (verbosity>0) fclose(logf);
  return(0);
} /* main */
